import struct


class Message(object):
    def __init__(self, data='', crc32=0, encoding='ascii', endian='big'):
        self.__data = data
        self.__crc32 = crc32
        self.__encoding = encoding
        self.__endian = endian
        self.__message = None

    @property
    def message(self):
        return self.__message

    @property
    def data(self):
        return self.__data

    @data.setter
    def data(self, new_data):
        self.__data = new_data

    @property
    def crc32(self):
        return self.__crc32

    @crc32.setter
    def crc32(self, crc):
        self.__crc32 = crc

    def __get_endian(self):
        if self.__endian.upper() == 'BIG':
            return '>'
        elif self.__endian.upper() == 'LITTLE':
            return '<'
        elif self.__endian.upper() == 'NATIVE':
            return '='
        elif self.__endian.upper() == 'NETWORK':
            return '!'
        else:
            return ''

    def encode(self, size: int) -> bytes:
        return struct.pack(self.__get_endian() + 'I', size)

    def decode(self, size_bytes: bytes) -> int:
        return struct.unpack(self.__get_endian() + 'I', size_bytes)[0]

    def encapsulate_msg(self):
        data = self.__data.encode(self.__encoding)
        data_length = self.encode(len(data))
        crc32 = self.encode(self.__crc32)
        message_length = self.encode(len(data) + len(data_length) + len(crc32))
        message = message_length + data_length + data + crc32
        self.__message = message

    def read_msg(self, msg_bytes: bytes):
        msg_len = self.decode(msg_bytes[:4])
        data_len = self.decode(msg_bytes[4:8])
        data = (msg_bytes[8:data_len+8]).decode(self.__encoding)
        crc32 = self.decode(msg_bytes[(data_len+8):])
        self.__crc32 = crc32
        self.__data = data



from array import array
from npipes.consts import *
from npipes.pipes import Pipes
from npipes.message import Message


class IPC(Pipes):
    def __init__(self, pipe_name=IPC_FIFO_NAME):
        super().__init__()
        self.__msg_parser = Message()
        self.__msg = ''
        self.pipe = pipe_name

    @property
    def message(self):
        return self.__msg

    @staticmethod
    def __calc_crc(msg, polynomial=0xEDB88320):
        # Ref: https://stackoverflow.com/questions/41564890/crc32-calculation-in-python-without-using-libraries
        table = array('L')
        for byte in range(256):
            crc = 0
            for bit in range(8):
                if (byte ^ crc) & 1:
                    crc = (crc >> 1) ^ polynomial
                else:
                    crc >>= 1
                byte >>= 1
            table.append(crc)
        value = 0xffffffff
        for ch in msg:
            value = table[(ord(ch) ^ value) & 0xff] ^ (value >> 8)
        return (-1 - value) & 0xffffffff

    def send_message(self, data: str, crc: bool = False) -> bool:
        try:
            crc32 = 0
            if crc:
                crc32 = IPC.__calc_crc(data)
            self.__msg_parser.data = data
            self.__msg_parser.crc32 = crc32
            self.__msg_parser.encapsulate_msg()
            self.__msg = self.__msg_parser.message
            self.write(self.__msg)
            return True
        except Exception as err:
            print(err)
            return False

    def receive_message(self, crc: bool = False) -> str:
        try:
            msg = self.read()
            self.__msg_parser.read_msg(msg)
            data = self.__msg_parser.data
            recv_crc32 = self.__msg_parser.crc32
            crc32 = IPC.__calc_crc(data)
            if crc:
                if recv_crc32 == crc32:
                    self.__msg = data
                    return data
                else:
                    print('Failed to validate the CRC')
                    print(f'Received CRC: {recv_crc32}')
                    print(f'Calculated CRC: {crc32}')
                    self.__msg = ''
                    return ''
            else:
                self.__msg = data
                return data
        except Exception as err:
            print(err)
            return ''






IPC_FIFO_NAME = 'IPC_PIPE_DEFAULT'
IPC_FIFO_NAME1 = 'IPC_PIPE_CUSTOM1'
IPC_FIFO_NAME2 = 'IPC_PIPE_CUSTOM2'
IPC_FIFO_NAME3 = 'IPC_PIPE_CUSTOM3'
IPC_FIFO_NAME4 = 'IPC_PIPE_CUSTOM4'
IPC_FIFO_NAME5 = 'IPC_PIPE_CUSTOM5'




from setuptools import setup, find_packages

setup(
    name='npipes',
    version='1.0.0',
    license='GPL-3.0',
    description='Wrapper Library for Named Pipes',
    author='Manikanta Ambadipudi',
    author_email='ambadipudi.manikanta@gmail.com',
    url='https://github.com/mani-src/npipes.git',
    download_url='https://github.com/mani-src/npipes/archive/v1.0.tar.gz',
    install_requires=['pywin32'],
    packages=find_packages()
)


import os
import time
import select
import win32pipe
import win32file
import pywintypes
# from win32security import CreateWellKnownSid, ACL, WinNetworkSid
from win32security import *

from npipes.consts import *
from npipes.message import Message


class Pipes(object):
    def __init__(self):
        self.__fifo = None
        self.__msg = None
        self.__pipe = IPC_FIFO_NAME

    @property
    def fifo(self):
        return self.__fifo

    @property
    def pipe(self):
        return self.__pipe

    @pipe.setter
    def pipe(self, pipe_name):
        self.__pipe = pipe_name

    def open(self, mode='read', pipe_wait=True):
        wait = win32pipe.PIPE_WAIT
        if pipe_wait:
            wait = win32pipe.PIPE_WAIT
        else:
            wait = win32pipe.PIPE_NOWAIT
        create_file = False
        try:
            self.__fifo = win32pipe.CreateNamedPipe(rf'\\.\pipe\{self.__pipe}', win32pipe.PIPE_ACCESS_DUPLEX,
                                                    win32pipe.PIPE_TYPE_MESSAGE | win32pipe.PIPE_READMODE_MESSAGE
                                                    | wait, 1, 65536, 65536, 0, None)
            print('Waiting for the client to connect')
            win32pipe.ConnectNamedPipe(self.__fifo, None)
            print('Client connected')
        except pywintypes.error as err:
            if mode == 'read':
                if err.args[0] == 231:
                    create_file = True
                else:
                    raise
            else:
                raise
        if mode == 'read' and create_file:
            self.__fifo = win32file.CreateFile(rf'\\.\pipe\{self.__pipe}',
                                               win32file.GENERIC_READ | win32file.GENERIC_WRITE, 0, None,
                                               win32file.OPEN_EXISTING, win32file.FILE_ATTRIBUTE_NORMAL, None)
            result = win32pipe.SetNamedPipeHandleState(self.__fifo, win32pipe.PIPE_READMODE_MESSAGE, None, None)
            if result == 0:
                raise Exception('Could not connect to any PIPE')
        return True

    def close(self):
        if self.__fifo:
            win32file.CloseHandle(self.__fifo)
            self.__fifo = None

    def write(self, msg):
        try:
            if not self.__fifo:
                self.open(mode='write')
            print(f'Sending message {msg} to the client')
            win32file.WriteFile(self.__fifo, msg)
            print(f'Message sent successfully')
        except Exception as exp:
            print(exp)
        finally:
            self.close()

    def read(self):
        try:
            if not self.__fifo:
                self.open()
            print(f'Reading message from the remote application')
            resp = win32file.ReadFile(self.__fifo, 65536)
            print(f'Message received: {resp}')
            return resp[1]
        except Exception as exp:
            print(exp)
        except pywintypes.error as err:
            if err.args[0] == 2:
                print('No PIPE available. Waiting for the PIPE to be created')
                time.sleep(1)
            elif err.args[0] == 109:
                print('Broken PIPE, exiting now')
            elif err.args[0] == 231:
                print('PIPE busy, check if the PIPE was opened successfully on the other end')
            print(err)
            return ''
        finally:
            self.close()

